#include <stdlib.h>
#include <limits.h>
#include <time.h>
#include <stdio.h>
#include <string.h>

#include "Construtivo.h"
#include "Global.h"

#define INCLUIR_LP(E, C, N) { tpNo _T = C[ N ] ; C[ N ] = C[ E ] ; C[ E ] = _T ; E = N++ ; }
#define REMOVER_LP(E, C, N) { tpNo _T = C[ N ] ; C[ N ] = C[ E ] ; C[ E ] = _T ; E = N-- ; }

tpCusto VizinhoMaisProximo( tpGrafo Grafo   ,
                            tpNo    NumNos  ,
                            tpNo    No      ,
                            double  Alfa    ,
                            tpNo**  Memoria  ) {

   tpNo    i                    ;
   tpNo    Cand                 ;
   tpNo    DivLP = 0            ; // Diviso da lista de pertinncias.
   tpCusto Menor                ;
   tpCusto LimSup               ;
   tpCusto Total = 0            ;
   tpNo*   Adj   = Memoria[ 0 ] ; // Vetor de adjacncias.
   tpNo*   LP    = Memoria[ 1 ] ; // Lista de pertinncia.
   tpNo*   LC    = Memoria[ 2 ] ; // Lista de candidatos.

   if ( NumNos < 1 ) {
      return 0 ;
   } /* if */

   // Inicializar a lista de pertinncia com todos os ns.
   for ( i = 0 ; i < NumNos ; i++ ) {
      LP[ i ] = i ;
   } /* for */

   Adj[ No ] = No ;
   INCLUIR_LP( No, LP, DivLP ) ;

   while ( DivLP < NumNos ) {

      // *** ALEATORIAZAO ***
      Cand = -1 ;
      Menor = INT_MAX ;
      LimSup = -1 ;

      for ( i = DivLP ; i < NumNos ; i++ ) {
         if ( ( PESO( Grafo, LP[ No ], LP[ i ] ) < INT_MAX ) &&
              ( PESO( Grafo, LP[ No ], LP[ i ] ) < Menor ) ) {

            Menor = PESO( Grafo, LP[ No ], LP[ i ] ) ;

         }
         if ( ( PESO( Grafo, LP[ No ], LP[ i ] ) < INT_MAX ) &&
              ( PESO( Grafo, LP[ No ], LP[ i ] ) > LimSup ) ) {
            
            LimSup = PESO( Grafo, LP[ No ], LP[ i ] ) ;

         }
      } /* for */
      if ( LimSup >= Menor ) {
         LimSup = (int)( Menor + Alfa * ( LimSup - Menor ) ) ;
         Cand = 0 ;
         for ( i = DivLP ; i < NumNos ; i++ ) {
            if ( PESO( Grafo, LP[ No ], LP[ i ] ) <= LimSup ) {
               LC[ Cand++ ] = i ;
            } /* if */
         } /* for */
         Cand = LC[ (int)RAND( Cand ) ] ;
      }
      // *** FIM ALEATORIAZAO ***

      if ( Cand >= 0 ) {
         Adj[ LP[ Cand ] ] = Adj[ LP[ No ] ] ;
         Adj[ LP[ No ] ] = LP[ Cand ] ;
         Total += PESO( Grafo, LP[ No ], LP[ Cand ] ) ;
         No = Cand ;
         INCLUIR_LP( No, LP, DivLP ) ;
      }
      else return -1;

   } /* while */

   // Verifica se  possvel fechar o ciclo.
   return PESO( Grafo, LP[ 0 ], LP[ No ] ) < INT_MAX ?
                              Total + PESO( Grafo, LP[ 0 ], LP[ No ] ) : -1 ;

} /* VizinhoMaisProximo */


tpCusto InsercaoMaisProxima( tpGrafo Grafo   ,
                             tpNo    NumNos  ,
                             tpNo    No      ,
                             double  Alfa    ,
                             tpNo**  Memoria  ) {

   tpCusto Custo                ;
   tpCusto Menor                ;
   tpCusto LimSup               ;
   tpCusto Total = 0            ;
   tpNo    i                    ;
   tpNo    Cand                 ;
   tpNo    DivLP = 0            ; // Diviso da lista de pertinncias.
   tpNo*   AD    = Memoria[ 0 ] ; // Vetor de adjacncias e distncias.
   tpNo*   LP    = Memoria[ 1 ] ; // Lista de pertinncia.
   tpNo*   LC    = Memoria[ 2 ] ; // Lista de candidatos.

   // Inicializar a lista de pertinncia com todos os ns e o vetor de
   // adjacncias sem ciclo.
   for ( i = 0 ; i < NumNos ; i++ ) {
      LP[ i ] = i       ;
      AD[ i ] = INT_MAX ;
   } /* for */

   AD[ No ] = No ;
   INCLUIR_LP( No, LP, DivLP ) ;
   
   while ( DivLP < NumNos ) {

      // Atualizar a lista de distncias e encontrar o vrtice fora do ciclo
      // mais distante.
      // *** ALEATORIAZAO ***
      Cand   = -1      ;
      Menor  = INT_MAX ;
      LimSup = -1      ;

      for ( i = DivLP ; i < NumNos ; i++ ) {
         Custo = PESO( Grafo, LP[ No ], LP[ i ] ) ;
         if ( Custo < AD[ LP[ i ] ] ) {
            AD[ LP[ i ] ] = Custo ;
         } /* if */
         if ( ( AD[ LP[ i ] ] < INT_MAX ) && ( AD[ LP[ i ] ] < Menor ) ) {
            Menor = AD[ LP[ i ] ] ;
         } /* if */
         if ( ( AD[ LP[ i ] ] < INT_MAX ) && ( AD[ LP[ i ] ] > LimSup ) ) {
            LimSup = AD[ LP[ i ] ] ;
         } /* if */
      } /* for */
      if ( Menor <= LimSup ) {
         LimSup = (int)( Menor + Alfa * ( LimSup - Menor ) ) ;
         Cand = 0 ;
         for ( i = DivLP ; i < NumNos ; i++ ) {
            if ( AD[ LP[ i ] ] <= LimSup ) {
               LC[ Cand++ ] = i ;
            } /* if */
         } /* for */
         Cand = LC[ (int)RAND( Cand ) ] ;
      } /* if */
      // *** FIM ALEATORIAZAO ***

      if ( Cand >= 0 ) {

         No    = Cand    ;
         Cand  = -1      ;
         Menor = INT_MAX ;

         // Calcular insero mais barata.
         i = LP[ 0 ] ;
         do {
            Custo = PESO( Grafo, i, LP[ No ] ) +
                    PESO( Grafo, LP[ No ], AD[ i ] ) ;
            
            // Corrgir overflow.
            if ( Custo < 0 ) Custo = INT_MAX ;
            else Custo -= PESO( Grafo, i, AD[ i ] ) ;
            
            if ( Custo < Menor ) {
               Menor = Custo ;
               Cand = i ;
            } /* if */

            i = AD[ i ] ;
         } while ( i != LP[ 0 ] ) ; /* do */

         if ( Cand >= 0 ) {
            // Inserir vrtice encontrado no ciclo.
            AD[ LP[ No ] ] = AD[ Cand ] ;
            AD[ Cand ] = LP[ No ] ;
            INCLUIR_LP( No, LP, DivLP ) ;
            Total += Menor ;
         } /* if */
         else return -1;

      } /* if */
      else return -1;

   } /* while */

   return Total ;

} /* InsercaoMaisProxima */


tpCusto InsercaoMaisDistante( tpGrafo Grafo   ,
                               tpNo    NumNos ,
                               tpNo    No     ,
                               double  Alfa   ,
                               tpNo**  Memoria  ) {

   tpCusto  Custo               ;
   tpCusto  Maior               ;
   tpCusto  LimInf              ;
   tpCusto  Total = 0           ;
   tpNo     i                   ;
   tpNo     Cand                ;
   tpNo     DivLP = 0           ; // Diviso da lista de pertinncias.
   tpNo*   AD    = Memoria[ 0 ] ; // Vetor de adjacncias e distncias.
   tpNo*   LP    = Memoria[ 1 ] ; // Lista de pertinncia.
   tpNo*   LC    = Memoria[ 2 ] ; // Lista de candidatos.

   // Inicializar a lista de pertinncia com todos os ns e o vetor de
   // adjacncias sem ciclo.
   for ( i = 0 ; i < NumNos ; i++ ) {
      LP[ i ] = i       ;
      AD[ i ] = INT_MAX ;
   } /* for */

   AD[ No ] = No ;
   INCLUIR_LP( No, LP, DivLP ) ;
   
   while ( DivLP < NumNos ) {

      // Atualizar a lista de distncias e encontrar o vrtice fora do ciclo
      // mais distante.
      // *** ALEATORIAZAO ***
      Cand   = -1      ;
      Maior  = -1      ;
      LimInf = INT_MAX ;

      for ( i = DivLP ; i < NumNos ; i++ ) {
         Custo = PESO( Grafo, LP[ No ], LP[ i ] ) ;
         if ( Custo < AD[ LP[ i ] ] ) {
            AD[ LP[ i ] ] = Custo ;
         } /* if */
         if ( ( AD[ LP[ i ] ] < INT_MAX ) && ( AD[ LP[ i ] ] < LimInf ) ) {
            LimInf = AD[ LP[ i ] ] ;
         } /* if */
         if ( ( AD[ LP[ i ] ] < INT_MAX ) && ( AD[ LP[ i ] ] > Maior ) ) {
            Maior = AD[ LP[ i ] ] ;
         } /* if */
      } /* for */
      if ( LimInf <= Maior ) {
         LimInf = (int)( Maior - Alfa * ( Maior - LimInf ) ) ;
         Cand = 0 ;
         for ( i = DivLP ; i < NumNos ; i++ ) {
            if ( AD[ LP[ i ] ] >= LimInf ) {
               LC[ Cand++ ] = i ;
            } /* if */
         } /* for */
         Cand = LC[ (int)RAND( Cand ) ] ;
      } /* if */
      // *** FIM ALEATORIAZAO ***

      if ( Cand >= 0 ) {

         No         = Cand;
         Cand  = -1        ;
         Custo = INT_MAX   ;

         // Calcular insero mais barata.
         i = LP[ 0 ] ;
         do {
            Maior = PESO( Grafo, i, LP[ No ] ) +
                          PESO( Grafo, LP[ No ], AD[ i ] ) ;
            
            // Corrgir overflow.
            if ( Maior < 0 ) Maior = INT_MAX ;
            else Maior -= PESO( Grafo, i, AD[ i ] ) ;
            
            if ( Maior < Custo ) {
               Custo = Maior ;
               Cand = i ;
            } /* if */

            i = AD[ i ] ;
         } while ( i != LP[ 0 ] ) ; /* do */

         if ( Cand >= 0 ) {
            // Inserir vrtice encontrado no ciclo.
            AD[ LP[ No ] ] = AD[ Cand ] ;
            AD[ Cand ] = LP[ No ] ;
            INCLUIR_LP( No, LP, DivLP ) ;
            Total += Custo ;
         } /* if */
         else return -1;

      } /* if */
      else return -1;

   } /* while */

   return Total ;

} /* InsercaoMaisDistante */


tpCusto InsercaoMaisBarata( tpGrafo Grafo   ,
                             tpNo    NumNos ,
                             tpNo    No     ,
                             double  Alfa   ,
                             tpNo**  Memoria  ) {

   tpCusto Custo                ,
           Menor                ,
           Total  = 0           ;
   tpNo    i                    ,
           j                    ,
           Aresta = 0           ,
           DivLP  = 0           ; // Diviso da lista de pertinncias.
   tpNo*   AI    = Memoria[ 0 ] ; // Vetor de adjacncias e distncias.
   tpNo*   LP    = Memoria[ 1 ] ; // Lista de pertinncia.
   tpNo*   LC    = Memoria[ 2 ] ; // Lista de candidatos (guarda as
                                  // economias temporariamente).

   // Inicializar a lista de pertinncia com todos os ns e o vetor de
   // adjacncias sem ciclo.
   for ( i = 0 ; i < NumNos ; i++ ) {
      LP[ i ] = i  ;
      AI[ i ] = -1 ;
   } /* for */

   AI[ No ] = No ;
   INCLUIR_LP( No, LP, DivLP ) ;
   
   while ( DivLP < NumNos ) {

      // Encontrar o vrtice fora do ciclo cuja insero  mais barata.
      for ( i = DivLP ; i < NumNos ; i++ ) {

         Menor = INT_MAX ;
         j = LP[ 0 ] ;
         do {
            
            Custo = PESO( Grafo, j, LP[ i ] ) +
                    PESO( Grafo, LP[ i ], AI[ j ] ) ;

            
            
            // Corrgir overflow.
            if ( Custo < 0 ) Custo = INT_MAX ;
            else             Custo -= PESO( Grafo, j, AI[ j ] ) ;
            
            if ( Custo < Menor ) {
               Menor = Custo ;
               Aresta = j ;
            } /* if */

            j = AI[ j ] ;
         } while ( j != LP[ 0 ] ) ; /* do */

         AI[ LP[ i ] ] = Aresta ;
         LC[ i ] = Menor ;
      } /* for */

      // *** ALEATORIAZAO ***
      No     = -1      ;
      Aresta = -1      ;
      Menor  = INT_MAX ;
      Custo  = -1      ;

      for ( i = DivLP ; i < NumNos ; i++ ) {
         if ( ( LC[ i ] < INT_MAX ) && ( LC[ i ] < Menor ) ) {
            Menor = LC[ i ] ;
         } /* if */
         if ( ( LC[ i ] < INT_MAX ) && ( LC[ i ] > Custo ) ) {
            Custo = LC[ i ] ;
         } /* if */
      } /* for */
      if ( Menor <= Custo ) {
         Custo = (int)( Menor + Alfa * ( Custo - Menor ) ) ;
         No = 0 ;
         for ( i = DivLP ; i < NumNos ; i++ ) {
            if ( LC[ i ] <= Custo ) {
               LC[ No++ ] = i ;
            } /* if */
         } /* for */
         No = LC[ (int)RAND( No ) ] ;
         Aresta = AI[ LP[ No ] ] ;
      } /* if */
      // *** FIM ALEATORIAZAO ***

      if ( Aresta >= 0 ) {
         // Inserir vrtice encontrado no ciclo.
         Total += PESO( Grafo, Aresta, LP[ No ] ) +
                  PESO( Grafo, LP[ No ], AI[ Aresta ] ) -
                  PESO( Grafo, Aresta, AI[ Aresta ] ) ;
         AI[ LP[ No ] ] = AI[ Aresta ] ;
         AI[ Aresta ] = LP[ No ] ;
         INCLUIR_LP( No, LP, DivLP ) ;
      } /* if */
      else return -1;

   } /* while */

   return Total ;

} /* InsercaoMaisBarata */


void AnalisarCons( tpMetodoConst* pAlgoritmo  ,
                   tpNo           MaxNos      ,
                   char*          szTipo      ,
                   char*          szAlgoritmo  )  {

   tpNo    i             ,
           j             ;
   tpCusto Custo_1       ,
           Custo_2       ,
           Custo_3       ;
   double  dTempo        ;
   clock_t TempoInicial  ,
           TempoFinal    ;
   tpGrafo Grafo_1       ,
           Grafo_2       ,
           Grafo_3       ; 
   tpNo*   pMemoria[ 3 ] ;
   
   printf( szAlgoritmo ) ;
   printf( szTipo ) ;
   
   for ( i = 0 ; i < 3 ; i++ ) {
      pMemoria[ i ] = ( tpNo* )malloc( sizeof( tpNo ) * MaxNos ) ;
      if ( pMemoria == NULL ) {
         printf( SZ_ERR_MEM ) ;
         exit( 0 ) ;
      } /* if */  
   }

   for ( i = 1 ; i <= MaxNos ; i += 20 ) {
      printf( "\n%d ", i ) ;
      Grafo_1 = AlocarGrafo( i ) ;
      Grafo_2 = AlocarGrafo( i ) ;
      Grafo_3 = AlocarGrafo( i ) ;
      Custo_1 = INT_MAX ;
      Custo_2 = INT_MAX ;
      Custo_3 = INT_MAX ;

      if ( strcmp( szTipo, "aleatorio" ) == 0 ) {
         CriarGrafo( Grafo_1, i, 1. ) ;
         CriarGrafo( Grafo_2, i, 1. ) ;
         CriarGrafo( Grafo_3, i, 1. ) ;
      } else {
         CriarGrafoEuclidiano( Grafo_1, i, 1., pMemoria ) ;
         CriarGrafoEuclidiano( Grafo_2, i, 1., pMemoria ) ;
         CriarGrafoEuclidiano( Grafo_3, i, 1., pMemoria ) ;
      } /* if */

      TempoInicial = clock( ) ;
      for ( j = 0 ; j < i ; j++ ) {
         Custo_1 = __min( Custo_1, pAlgoritmo( Grafo_1, i, j, 0., pMemoria ) ) ;
         Custo_2 = __min( Custo_2, pAlgoritmo( Grafo_2, i, j, 0., pMemoria ) ) ;
         Custo_3 = __min( Custo_3, pAlgoritmo( Grafo_3, i, j, 0., pMemoria ) ) ;
      } /* for */
      TempoFinal = clock( ) ;

      DestruirGrafo( Grafo_1, i ) ;
      DestruirGrafo( Grafo_2, i ) ;
      DestruirGrafo( Grafo_3, i ) ;
     
      dTempo = ( double )( TempoFinal - TempoInicial ) / ( CLOCKS_PER_SEC * 3. ) ;
      printf( "\t%lf\t%.10lf", ( double )( Custo_1 + Custo_2 + Custo_3 ) / 3.,
                                                               dTempo ) ;
      fflush( stdout ) ;
   } /* for */

   for ( i = 0 ; i < 3 ; i++ ) {
      free( pMemoria[ i ] ) ;
   } /* for */
   
} /* AnalisarCons */


void AnalisarConsInst( tpGrafo        Grafo       ,
                       tpNo           NumNos      ,
                       tpMetodoConst* pAlgoritmo  ,
                       char*          szAlgoritmo ,
                       tpNo**         pMemoria    ,
                       tpCusto*       CustoMin    ,
                       char*          szMaisEfic   ) {

   tpCusto Custo = *CustoMin ;
   int     i                 ;

   for ( i = 0 ; i < NumNos ; i++ ) {
      Custo = __min( Custo, pAlgoritmo( Grafo, NumNos, i, 0, pMemoria ) ) ;
   } /* for */
   
   if ( Custo < *CustoMin ) {
      *CustoMin = Custo ;
      strcpy( szMaisEfic, szAlgoritmo ) ;
   } /* if */
   
} /* AnalisarConsInst */     