#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#include "GRASP.h"
#include "Global.h"


tpCusto GRASP( tpGrafo       Grafo     ,
               tpNo          NumNos    ,
               tpMetodoConst Metodo    ,
               tpBuscaLocal  Busca     ,
               double        Alfa      ,
               int           Iteracoes ,
               tpNo**        Memoria    ) {

   tpNo    No                      ;
   tpCusto Custo                   ,
           CustoIni     = INT_MAX  ,
           Melhor       = INT_MAX  ;
   tpNo*   Tempo                   ;
   clock_t TempoInicial = clock( ) ,
           TempoFinal              ;
   double  dTempo                  ;

   while ( Iteracoes > 0 ) {

      No = (int)RAND( NumNos ) ;
      
      Custo = Metodo( Grafo, NumNos, No, Alfa, Memoria ) ;
      Custo = Busca( Grafo, Custo, Memoria[ 0 ] ) ;
      
      if ( CustoIni == INT_MAX ) {
         CustoIni = Custo ;
      } /* if */

      if ( Custo < Melhor ) {
         TempoFinal = clock( ) ;
         dTempo = ( double )( TempoFinal - TempoInicial ) / CLOCKS_PER_SEC ;
         printf( "\t%lf\t%.10lf", 1. - ( ( double )Custo / ( double )CustoIni ), dTempo ) ;
         fflush( stdout ) ;

         Melhor = Custo ;
         Tempo = Memoria[ 3 ] ;
         Memoria[ 3 ] = Memoria[ 0 ] ;
         Memoria[ 0 ] = Tempo ;
      } /* if */

      Iteracoes-- ;
   }

   Tempo = Memoria[ 3 ] ;
   Memoria[ 3 ] = Memoria[ 0 ] ;
   Memoria[ 0 ] = Tempo ;

   return Melhor ;

} /* GRASP */


void AnalisarGRASP( tpMetodoConst* pAlgoritmo  ,
                    tpBuscaLocal*  pBusca      , 
                    char*          szTipo      ,
                    char*          szAlgoritmo ,
                    char*          szBusca      ) {

   tpNo    i             ;
   tpCusto Custo         ;
   tpGrafo Grafo         ;  
   tpNo*   pMemoria[ 4 ] ;
   
   printf( szAlgoritmo ) ;
   printf( szBusca ) ;
   printf( szTipo ) ;

   for ( i = 0 ; i < 4 ; i++ ) {
      pMemoria[ i ] = ( tpNo* )malloc( sizeof( tpNo ) * 100 ) ;
      if ( pMemoria == NULL ) {
         printf( SZ_ERR_MEM ) ;
         exit( 0 ) ;
      } /* if */  
   } 

   for ( i = 50 ; i <= 100 ; i += 50 ) {
      printf( "\n%d ", i ) ;
      Grafo = AlocarGrafo( i ) ;

      if ( strcmp( szTipo, "aleatorio" ) == 0 ) {
         CriarGrafo( Grafo, i, 1. ) ;
      } else {
         CriarGrafoEuclidiano( Grafo, i, 1., pMemoria ) ;
      } /* if */

      Custo = GRASP( Grafo, i, *pAlgoritmo, *pBusca, 0.1, 500, pMemoria ) ;  
      printf( "\t%d", Custo ) ;
      DestruirGrafo( Grafo, i ) ;
   } /* for */

   for ( i = 0 ; i < 4 ; i++ ) {
      free( pMemoria[ i ] ) ;
   } /* for */
   
} /* AnalisarGRASP */ 


void AnalisarGRASPInst( tpGrafo        Grafo       ,
                        tpNo           NumNos      ,
                        tpMetodoConst* pAlgoritmo  ,
                        tpBuscaLocal*  pBusca      , 
                        char*          szAlgoritmo ,
                        char*          szBusca     ,
                        tpNo**         pMemoria    ,
                        tpCusto*       CustoMin    ,
                        char*          szMaisEfic   ) {

   tpCusto Custo = GRASP( Grafo, NumNos, *pAlgoritmo, *pBusca, 0.1, 500, pMemoria ) ;
   
   if ( Custo < *CustoMin ) {
      *CustoMin = Custo ;
      strcpy( szMaisEfic, szAlgoritmo ) ;
      strcat( szMaisEfic, " com " ) ;
      strcat( szMaisEfic, szBusca ) ;
   } /* if */
   
} /* AnalisarGRASPInst */     