#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <math.h>
#include <string.h>

#include "Global.h"
#include "Grafo.h"

#define RAIZ_2            1.4142135623730950488016887242097

#define NINT(X)           fmod( X, 1 ) < 0.5 ? (int)floor( X ) : (int)ceil( X ) ;


tpGrafo AlocarGrafo( tpNo iNumNos ) {

   tpNo    i     ;
   tpGrafo Grafo ;
   
   // Alocao de memria.
   Grafo = ( tpGrafo )malloc( sizeof( int* ) * iNumNos ) ;
   if ( Grafo == NULL ) {
      printf( SZ_ERR_MEM ) ;
      exit( 0 ) ;
   } /* if */  

   // Alocao de memria dos vetores secundrios.
   for ( i = 0 ; i < iNumNos ; i++ ) {
      Grafo[ i ] = ( int* )malloc( sizeof( int ) * ( i + 1 ) ) ;
      if ( Grafo[ i ] == NULL ) {
         printf( SZ_ERR_MEM ) ;
         exit( 0 ) ;
      } /* if */  
   } /* for */

   return Grafo ;
}


void CriarGrafo( tpGrafo Grafo      ,
                 tpNo    iNumNos    , 
                 double  dDensidade  ) {

   tpNo    i     ,
           j     ;
   tpCusto Custo ;
   
   // Construo do grafo.
   for ( i = 0 ; i < iNumNos ; i++ ) {
      ColocarPeso( Grafo, i, i, 0 ) ; 
      for ( j = i + 1 ; j < iNumNos ; j++ ) {
         
         if ( RAND( 1 ) > dDensidade ) { 
            Custo = INT_MAX ;
         } else {                          
            Custo = ( int )RAND( CUSTO_MAX ) ;
         } /* if */
         
         ColocarPeso( Grafo, i, j, Custo ) ;
      } /* for */
   } /* for */
      
} /* CriarGrafo */


void CriarGrafoEuclidiano( tpGrafo Grafo      ,
                           tpNo    iNumNos    , 
                           double  dDensidade ,
                           tpNo**  Memoria     ) {

   tpNo    i                      ,
           j                      ,
           xd                     ,
           yd                     ;
   tpNo*   PosX    = Memoria[ 0 ] ;
   tpNo*   PosY    = Memoria[ 1 ] ;
   tpCusto Custo                  ;
   
   for ( i = 0 ; i < iNumNos ; i++ ) {
      PosX[ i ] = (int) RAND( CUSTO_MAX / RAIZ_2 ) ;
      PosY[ i ] = (int) RAND( CUSTO_MAX / RAIZ_2 ) ;
   }

   // Construo do grafo.
   for ( i = 0 ; i < iNumNos ; i++ ) {
      ColocarPeso( Grafo, i, i, 0 ) ;       
      for ( j = i + 1 ; j < iNumNos ; j++ ) {
         
         if ( RAND( 1 ) > dDensidade ) {
            Custo = INT_MAX ;
         } else {
            xd = PosX[ i ] - PosX[ j ] ;
            yd = PosY[ i ] - PosY[ j ] ;
            Custo = NINT( sqrt( xd * xd + yd * yd ) ) ;
         } /* if */
         
         ColocarPeso( Grafo, i, j, Custo ) ;
      } /* for */
   } /* for */
   
} /* CriarGrafoEuclidiano */


void DestruirGrafo( tpGrafo Grafo   , 
                    tpNo    iNumNos  ) { 
                   
   tpNo i ;

   for ( i = 0 ; i < iNumNos ; i++ ) {
      free( Grafo[ i ] ) ;
   } /* for */

   free( Grafo ) ;

} /* DestruirGrafo */


void ColocarPeso( tpGrafo Grafo     ,
                  tpNo    iIndice_i ,
                  tpNo    iIndice_j ,
                  tpCusto Custo      ) { 

   if ( iIndice_i < iIndice_j ) {
      Grafo[ iIndice_j ][ iIndice_i ] = Custo ;
   } else {
      Grafo[ iIndice_i ][ iIndice_j ] = Custo ;
   } /* if */

} /* ColocarPeso */    