#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "Grafo.h"
#include "Leitor.h"

#define KW_SINTAX "%s :"
#define PI        3.141592
#define RRR       6378.388

#define SECAO_COORD_NOS_FLAG     1
#define SECAO_DEPOT_FLAG         2
#define SECAO_DEMANDA_FLAG       4
#define SECAO_ARESTAS_FLAG       8
#define SECAO_ARESTAS_FIXAS_FLAG 16
#define SECAO_EXIBICAO_FLAG      32
#define SECAO_ROTA_FLAG          64
#define SECAO_CUSTOS_FLAG        128
#define FIM_FLAG                 256

#define NUM_KW         19
#define NUM_KW_COM_VAL 6

#define TIPO                0
#define TIPO_CUSTO          1
#define FORMATO_CUSTO       2
#define FORMATO_ARESTA      3
#define COORDENADAS         4
#define TIPO_EXIBICAO       5 
#define NOME                6
#define COMENTARIO          7
#define DIMENSAO            8
#define CAPACIDADE          9
#define SECAO_COORD_NOS     10
#define SECAO_DEPOT         11
#define SECAO_DEMANDA       12
#define SECAO_ARESTAS       13
#define SECAO_ARESTAS_FIXAS 14
#define SECAO_EXIBICAO      15
#define SECAO_ROTA          16
#define SECAO_CUSTOS        17
#define FIM                 18

#define SIMETRICO    0
#define ASSIMETRICO  1
#define ORDENAMENTO  2
#define HAMILTONIANO 3
#define ROTEAMENTO   4
#define ROTA         5

#define EXCPLICITO       0
#define COORD_2D         1
#define COORD_3D         2
#define DIST_MAX_2D      3
#define DIST_MAX_3D      4
#define DIST_MAN_2D      5
#define DIST_MAN_3D      6
#define COORD_EUC_ARREND 7
#define COORD_GEOGRAF    8
#define ESPECIAL_ATT     9
#define ESPECIAL_XRAY1   10
#define ESPECIAL_XRAY2   11

#define FUNCAO_CUSTO        0
#define MATRIZ_COMPLETA     1
#define MATRIZ_SUP_LIN      2
#define MATRIZ_INF_LIN      3
#define MATRIZ_SUP_DIAG_LIN 4
#define MATRIZ_INF_DIAG_LIN 5
#define MATRIZ_SUP_COL      6
#define MATRIZ_INF_COL      7
#define MATRIZ_SUP_DIAG_COL 8
#define MATRIZ_INF_DIAG_COL 10

#define LISTA_ADJACENCIAS   0
#define LISTA_ARESTAS       1

#define SEM_COORD 0
//#define COORD_2D  1
//#define COORD_3D  2

#define SEM_EXIB      0
#define EXIB_COORD    1
#define EXIB_2D       2

#define MAX(X, Y) ( X > Y ? X : Y )

int nint( double valor ) { return (int)( valor + 0.5 ) ; }

void CalcularCustos( tpInstancia Instancia, double** Coords, int TipoCusto ) {
   int	      i, j, t ;
   double  	   xd, yd, zd, r, q1, q2, q3, min, deg ;

   switch ( TipoCusto ) {
   case COORD_2D:
      for ( i = 0 ; i < Instancia->NumNos ; i++ ) {
         Instancia->Grafo[ i ][ i ] = 0 ;
         for ( j = 0 ; j < i ; j++ ) {
            xd = Coords[ 0 ][ i ] - Coords[ 0 ][ j ] ;
            yd = Coords[ 1 ][ i ] - Coords[ 1 ][ j ] ;
            Instancia->Grafo[ i ][ j ] = nint( sqrt( xd * xd + yd * yd ) ) ;
         }
      }
      break ;
   case COORD_3D:
      for ( i = 0 ; i < Instancia->NumNos ; i++ ) {
         Instancia->Grafo[ i ][ i ] = 0 ;
         for ( j = 0 ; j < i ; j++ ) {
            xd = Coords[ 0 ][ i ] - Coords[ 0 ][ j ] ;
            yd = Coords[ 1 ][ i ] - Coords[ 1 ][ j ] ;
            zd = Coords[ 2 ][ i ] - Coords[ 2 ][ j ] ;
            Instancia->Grafo[ i ][ j ] = nint( sqrt( xd * xd + yd * yd + zd * zd ) ) ;
         }
      }
      break ;
   case DIST_MAX_2D:
      for ( i = 0 ; i < Instancia->NumNos ; i++ ) {
         Instancia->Grafo[ i ][ i ] = 0 ;
         for ( j = 0 ; j < i ; j++ ) {
            xd = fabs( Coords[ 0 ][ i ] - Coords[ 0 ][ j ] ) ;
            yd = fabs( Coords[ 1 ][ i ] - Coords[ 1 ][ j ] ) ;
            Instancia->Grafo[ i ][ j ] = MAX( nint( xd ), nint( yd ) ) ;
         }
      }
      break ;
   case DIST_MAX_3D:
      for ( i = 0 ; i < Instancia->NumNos ; i++ ) {
         Instancia->Grafo[ i ][ i ] = 0 ;
         for ( j = 0 ; j < i ; j++ ) {
            xd = fabs( Coords[ 0 ][ i ] - Coords[ 0 ][ j ] ) ;
            yd = fabs( Coords[ 1 ][ i ] - Coords[ 1 ][ j ] ) ;
            zd = fabs( Coords[ 2 ][ i ] - Coords[ 2 ][ j ] ) ;
            Instancia->Grafo[ i ][ j ] = MAX( nint( xd ), MAX( nint( yd ), nint( zd ) ) ) ;
         }
      }
      break ;
   case DIST_MAN_2D:
      for ( i = 0 ; i < Instancia->NumNos ; i++ ) {
         Instancia->Grafo[ i ][ i ] = 0 ;
         for ( j = 0 ; j < i ; j++ ) {
            xd = fabs( Coords[ 0 ][ i ] - Coords[ 0 ][ j ] ) ;
            yd = fabs( Coords[ 1 ][ i ] - Coords[ 1 ][ j ] ) ;
            Instancia->Grafo[ i ][ j ] = nint( xd + yd ) ;
         }
      }
      break ;
   case DIST_MAN_3D:
      for ( i = 0 ; i < Instancia->NumNos ; i++ ) {
         Instancia->Grafo[ i ][ i ] = 0 ;
         for ( j = 0 ; j < i ; j++ ) {
            xd = fabs( Coords[ 0 ][ i ] - Coords[ 0 ][ j ] ) ;
            yd = fabs( Coords[ 1 ][ i ] - Coords[ 1 ][ j ] ) ;
            zd = fabs( Coords[ 2 ][ i ] - Coords[ 2 ][ j ] ) ;
            Instancia->Grafo[ i ][ j ] = nint( xd + yd + zd ) ;
         }
      }
      break ;
   case COORD_GEOGRAF:
      for ( i = 0 ; i < Instancia->NumNos ; i++ ) {
         deg = (int)Coords[ 0 ][ i ] ;
         min = Coords[ 0 ][ i ] - deg ;
         Coords[ 0 ][ i ] = PI * ( deg + 5.0 * min / 3.0 ) / 180.0 ;
         deg = (int)Coords[ 1 ][ i ] ;
         min = Coords[ 1 ][ i ] - deg ;
         Coords[ 1 ][ i ] = PI * ( deg + 5.0 * min / 3.0 ) / 180.0 ;
      }
      for ( i = 0 ; i < Instancia->NumNos ; i++ ) {
         Instancia->Grafo[ i ][ i ] = 0 ;
         for ( j = 0 ; j < i ; j++ ) {
            q1 = cos( Coords[ 1 ][ i ] - Coords[ 1 ][ j ] );
            q2 = cos( Coords[ 0 ][ i ] - Coords[ 0 ][ j ] );
            q3 = cos( Coords[ 0 ][ i ] + Coords[ 0 ][ j ] );

            Instancia->Grafo[ i ][ j ] = (int)( RRR * acos( 0.5*((1.0+q1)*q2 - (1.0-q1)*q3) ) + 1.0 ) ;
         }
      }
      break ;
   case ESPECIAL_ATT:
      for ( i = 0 ; i < Instancia->NumNos ; i++ ) {
         Instancia->Grafo[ i ][ i ] = 0 ;
         for ( j = 0 ; j < i ; j++ ) {
            xd = fabs( Coords[ 0 ][ i ] - Coords[ 0 ][ j ] ) ;
            yd = fabs( Coords[ 1 ][ i ] - Coords[ 1 ][ j ] ) ;
            r = sqrt( ( xd * xd + yd * yd ) / 10.0 ) ;
            t = nint( r ) ;
            if ( t < r ) Instancia->Grafo[ i ][ j ] = t + 1 ;
            else         Instancia->Grafo[ i ][ j ] = t ;
         }
      }
      break ;
   case COORD_EUC_ARREND:
      for ( i = 0 ; i < Instancia->NumNos ; i++ ) {
         Instancia->Grafo[ i ][ i ] = 0 ;
         for ( j = 0 ; j < i ; j++ ) {
            xd = Coords[ 0 ][ i ] - Coords[ 0 ][ j ] ;
            yd = Coords[ 1 ][ i ] - Coords[ 1 ][ j ] ;
            Instancia->Grafo[ i ][ j ] = (int)ceil( sqrt( xd * xd + yd * yd ) ) ;
         }
      }
      break ;
   default:
      printf( "Nenhum tipo de custo valido foi especificado." ) ;
      break;
   }
}

tpInstancia LerInstancia( char* Caminho ) {

   char*       KeyWords[ NUM_KW ]           ;
   char**      Valores[ NUM_KW_COM_VAL ]    ;
   int         NumValores[ NUM_KW_COM_VAL ] ;
   int         Secoes = 0                   ;
   int         i                            ;
   int         j                            ;
   int         KeyWord                      ;
   int         Valor                        ;
   int	      NumCoord = 0                 ;
   int         CamposLidos                  ;
   int         Espec[ NUM_KW_COM_VAL ]      ;
   tpNo        No                           ;
   tpNo        NosDepot = 0                 ;
   tpCusto     Custo                        ;
   tpInstancia Instancia                    ;
   FILE*       Arquivo                      ;
   double**    Coords                       ;
   double      Real                         ;
   char        Texto[ 255 ]                 ;
   int         Inteiro                      ;

   KeyWords[ NOME                ] = "NAME"                ;
   KeyWords[ TIPO                ] = "TYPE"                ;
   KeyWords[ COMENTARIO          ] = "COMMENT"             ;
   KeyWords[ DIMENSAO            ] = "DIMENSION"           ;
   KeyWords[ CAPACIDADE          ] = "CAPACITY"            ;
   KeyWords[ TIPO_CUSTO          ] = "EDGE_WEIGHT_TYPE"    ;
   KeyWords[ FORMATO_CUSTO       ] = "EDGE_WEIGHT_FORMAT"  ;
   KeyWords[ FORMATO_ARESTA      ] = "EDGE_DATA_FORMAT"    ;
   KeyWords[ COORDENADAS         ] = "NODE_COORD_TYPE"     ;
   KeyWords[ TIPO_EXIBICAO       ] = "DISPLAY_DATA_TYPE"   ;
   KeyWords[ SECAO_COORD_NOS     ] = "NODE_COORD_SECTION"  ;
   KeyWords[ SECAO_DEPOT         ] = "DEPOT_SECTION"       ;
   KeyWords[ SECAO_DEMANDA       ] = "DEMAND_SECTION"      ;
   KeyWords[ SECAO_ARESTAS       ] = "EDGE_DATA_SECTION"   ;
   KeyWords[ SECAO_ARESTAS_FIXAS ] = "FIXED_EDGES_SECTION" ;
   KeyWords[ SECAO_EXIBICAO      ] = "DISPLAY_DATA_SECTION"     ;
   KeyWords[ SECAO_ROTA          ] = "TOUR_SECTION"        ;
   KeyWords[ SECAO_CUSTOS        ] = "EDGE_WEIGHT_SECTION" ;
   KeyWords[ FIM                 ] = "EOF"                 ;

   NumValores[ TIPO           ] = 6  ;
   NumValores[ TIPO_CUSTO     ] = 12 ;
   NumValores[ FORMATO_CUSTO  ] = 11 ;
   NumValores[ FORMATO_ARESTA ] = 2  ;
   NumValores[ COORDENADAS    ] = 3  ;
   NumValores[ TIPO_EXIBICAO  ] = 3  ;

   Valores[ TIPO ]           = ( char** ) malloc( sizeof( char* ) * NumValores[ TIPO           ] ) ;
   if ( Valores[ TIPO ] == NULL ) {
      printf( SZ_ERR_MEM ) ;
      exit( 0 ) ;
   } /* if */ 

   Valores[ TIPO_CUSTO ]     = ( char** ) malloc( sizeof( char* ) * NumValores[ TIPO_CUSTO     ] ) ;
   if ( Valores[ TIPO_CUSTO ] == NULL ) {
      printf( SZ_ERR_MEM ) ;
      exit( 0 ) ;
   } /* if */ 

   Valores[ FORMATO_CUSTO ]  = ( char** ) malloc( sizeof( char* ) * NumValores[ FORMATO_CUSTO  ] ) ;
   if ( Valores[ FORMATO_CUSTO ] == NULL ) {
      printf( SZ_ERR_MEM ) ;
      exit( 0 ) ;
   } /* if */ 

   Valores[ FORMATO_ARESTA ] = ( char** ) malloc( sizeof( char* ) * NumValores[ FORMATO_ARESTA ] ) ;
   if ( Valores[ FORMATO_ARESTA ] == NULL ) {
      printf( SZ_ERR_MEM ) ;
      exit( 0 ) ;
   } /* if */ 

   Valores[ COORDENADAS ]    = ( char** ) malloc( sizeof( char* ) * NumValores[ COORDENADAS    ] ) ;
   if ( Valores[ COORDENADAS ] == NULL ) {
      printf( SZ_ERR_MEM ) ;
      exit( 0 ) ;
   } /* if */ 

   Valores[ TIPO_EXIBICAO ]  = ( char** ) malloc( sizeof( char* ) * NumValores[ TIPO_EXIBICAO  ] ) ;
   if ( Valores[ TIPO_EXIBICAO ] == NULL ) {
      printf( SZ_ERR_MEM ) ;
      exit( 0 ) ;
   } /* if */ 

   Valores[ TIPO ][ SIMETRICO    ] = "TSP"  ;
   Valores[ TIPO ][ ASSIMETRICO  ] = "ATSP" ;
   Valores[ TIPO ][ ORDENAMENTO  ] = "SOP"  ;
   Valores[ TIPO ][ HAMILTONIANO ] = "HCP"  ;
   Valores[ TIPO ][ ROTEAMENTO   ] = "CVRP" ;
   Valores[ TIPO ][ ROTA         ] = "TOUR" ;

   Valores[ TIPO_CUSTO ][ EXCPLICITO       ] = "EXPLICIT" ;
   Valores[ TIPO_CUSTO ][ COORD_2D         ] = "EUC_2D"   ;
   Valores[ TIPO_CUSTO ][ COORD_3D         ] = "EUC_3D"   ;
   Valores[ TIPO_CUSTO ][ DIST_MAX_2D      ] = "MAX_2D"   ;
   Valores[ TIPO_CUSTO ][ DIST_MAX_3D      ] = "MAX_3D"   ;
   Valores[ TIPO_CUSTO ][ DIST_MAN_2D      ] = "MAN_2D"   ;
   Valores[ TIPO_CUSTO ][ DIST_MAN_3D      ] = "MAN_3D"   ;
   Valores[ TIPO_CUSTO ][ COORD_EUC_ARREND ] = "CEIL_2D"  ;
   Valores[ TIPO_CUSTO ][ COORD_GEOGRAF    ] = "GEO"      ;
   Valores[ TIPO_CUSTO ][ ESPECIAL_ATT     ] = "ATT"      ;
   Valores[ TIPO_CUSTO ][ ESPECIAL_XRAY1   ] = "XRAY1"    ;
   Valores[ TIPO_CUSTO ][ ESPECIAL_XRAY2   ] = "XRAY2"    ;

   Valores[ FORMATO_CUSTO ][ FUNCAO_CUSTO        ] = "FUNCTION"       ;
   Valores[ FORMATO_CUSTO ][ MATRIZ_COMPLETA     ] = "FULL_MATRIX"    ;
   Valores[ FORMATO_CUSTO ][ MATRIZ_SUP_LIN      ] = "UPPER_ROW"      ;
   Valores[ FORMATO_CUSTO ][ MATRIZ_INF_LIN      ] = "LOWER_ROW"      ;
   Valores[ FORMATO_CUSTO ][ MATRIZ_SUP_DIAG_LIN ] = "UPPER_DIAG_ROW" ;
   Valores[ FORMATO_CUSTO ][ MATRIZ_INF_DIAG_LIN ] = "LOWER_DIAG_ROW" ;
   Valores[ FORMATO_CUSTO ][ MATRIZ_SUP_COL      ] = "UPPER_COL"      ;
   Valores[ FORMATO_CUSTO ][ MATRIZ_INF_COL      ] = "LOWER_COL"      ;
   Valores[ FORMATO_CUSTO ][ MATRIZ_SUP_DIAG_COL ] = "UPPER_DIAG_COL" ;
   Valores[ FORMATO_CUSTO ][ MATRIZ_INF_DIAG_COL ] = "LOWER_DIAG_COL" ;

   Valores[ FORMATO_ARESTA ][ LISTA_ADJACENCIAS ] = "ADJ_LIST"  ;
   Valores[ FORMATO_ARESTA ][ LISTA_ARESTAS     ] = "EDGE_LIST" ;

   Valores[ COORDENADAS ][ SEM_COORD ] = "NO_COORDS"     ;
   Valores[ COORDENADAS ][ COORD_2D  ] = "TWOD_COORDS"   ;
   Valores[ COORDENADAS ][ COORD_3D  ] = "THREED_COORDS" ;

   Valores[ TIPO_EXIBICAO ][ SEM_EXIB   ] = "NO_DISPLAY"     ;
   Valores[ TIPO_EXIBICAO ][ EXIB_COORD ] = "COORDS_DISPLAY" ;
   Valores[ TIPO_EXIBICAO ][ EXIB_2D    ] = "TWOD_DISPLAY"   ;   

   Instancia = ( tpInstancia )malloc( sizeof( struct tgInstancia ) ) ;
   if ( Instancia == NULL ) {
      printf( SZ_ERR_MEM ) ;
      exit( 0 ) ;
   } /* if */ 

   Instancia->NumNos = 0 ;
   Instancia->Grafo = NULL ;
   for ( KeyWord = 0 ; KeyWord < NUM_KW_COM_VAL ; KeyWord++ ) {
      Espec[ KeyWord ] = -1 ;
   }

   Arquivo = fopen( Caminho, "r" ) ;

   if ( Arquivo != NULL ) {

      do {
         CamposLidos = fscanf( Arquivo, KW_SINTAX, Texto ) ;

         if ( CamposLidos > 0 ) {
            for ( KeyWord = 0 ; KeyWord < NUM_KW ; KeyWord++ ) {
               if ( strcmp( Texto, KeyWords[ KeyWord ] ) == 0 ) {
                  if ( KeyWord < NUM_KW_COM_VAL ) {

                     CamposLidos = fscanf( Arquivo, "%s", Texto ) ;
                     if ( CamposLidos > 0 ) {
                        for ( Valor = 0 ; Valor < NumValores[ KeyWord ] ; Valor++ ) {
                           if ( strcmp( Texto, Valores[ KeyWord ][ Valor ] ) == 0 ) {
                              Espec[ KeyWord ] = Valor ;
                              break ;
                           }
                        }
                     }

                  } else {

                     switch ( KeyWord ) {
                     /////////////////////////////////////////////////////////////
                     case NOME:
                        CamposLidos = fscanf( Arquivo, "%s", Instancia->Nome ) ;
                        break ;
                     /////////////////////////////////////////////////////////////
                     case COMENTARIO:
                        Valor = 0 ;
                        while ( ( Instancia->Comentario[ Valor++ ] = ( char )fgetc( Arquivo ) ) != '\n' ) {} ;
                        Instancia->Comentario[ --Valor ] = '\0' ;
                        break ;
                     /////////////////////////////////////////////////////////////
                     case DIMENSAO:
                        CamposLidos = fscanf( Arquivo, "%d", &Inteiro ) ;
                        if ( CamposLidos > 0 ) {
                           Instancia->NumNos = Inteiro ;
                           Instancia->Grafo = AlocarGrafo( Instancia->NumNos ) ;
                        }
                        break ;
                     /////////////////////////////////////////////////////////////
                     case CAPACIDADE:
                        CamposLidos = fscanf( Arquivo, "%d", &Inteiro ) ;
                        if ( CamposLidos > 0 ) {
                           // S interessa ao problema de CVRP ;
                        } else {
                           printf( "Nao foi possivel ler a capacidade especificada.\n" ) ;
                           return NULL ;
                        }
                        break ;
                     /////////////////////////////////////////////////////////////
                     case SECAO_COORD_NOS:
                        Secoes = Secoes | SECAO_COORD_NOS_FLAG ; 

                        // Descobre o formato que as coordenadas esto representadas.
                        switch ( Espec[ COORDENADAS ] ) {
                        case SEM_COORD:
                           printf( "Instancia definida sem coordenadas, mas foi encontrada a secao de coordenadas de nos.\n" ) ;
                           return NULL ;
                        case COORD_2D:
                           NumCoord = 2;
                           break ;
                        case COORD_3D:
                           NumCoord = 3;
                           break ;
                        default:
                           switch ( Espec[ TIPO_CUSTO ] ) {
                           case COORD_2D:
                           case COORD_EUC_ARREND:
                           case COORD_GEOGRAF:
                           case ESPECIAL_ATT:
                              NumCoord = 2;
                              break ;
                           case COORD_3D:
                              NumCoord = 3;
                              break ;
                           default:
                              printf( "Nao foi possivel determinar o formato das coordenadas de nos fornecidas.\n" ) ;
                              return NULL ;
                           }
                           break ;
                        }

                        // Aloca espao para armazenar as coordenadas.
                        if ( NumCoord > 0 ) {
                           Coords = ( double** ) malloc( sizeof( double* ) * NumCoord ) ;
                           if ( Coords == NULL ) {
                              printf( SZ_ERR_MEM ) ;
                              exit( 0 ) ;
                           } /* if */ 
                           for ( i = 0 ; i < NumCoord ; i++ ) {
                              Coords[ i ] = ( double* ) malloc( sizeof( double ) * Instancia->NumNos ) ;
                              if ( Coords[ i ] == NULL ) {
                                 printf( SZ_ERR_MEM ) ;
                                 exit( 0 ) ;
                              } /* if */ 
                           }
                        }

                        // L as coordenadas do arquivo.
                        No = 0 ;
                        while ( No < Instancia->NumNos ) {
                           No++;
                           CamposLidos = fscanf( Arquivo, "%d", &Inteiro ) ;
                           if ( CamposLidos > 0 ) {
                              if ( ( Inteiro > 0 ) && ( Inteiro <= Instancia->NumNos ) ) {
                                 for ( i = 0 ; i < NumCoord ; i++ ) {
                                    CamposLidos = fscanf( Arquivo, "%lf", Coords[ i ] + Inteiro - 1 ) ;
                                    if ( CamposLidos <= 0 ) {
                                       printf( "Formato de especificacao de coordenadas de nos invalida [leitura do %do. no %da. coordenada].\n", No, i  ) ;
                                       return NULL ;
                                    }
                                 }
                              } else {
                                 printf( "Secao de coordenadas de nos com indices de nos fora do valor permitido [1; %d].\n", Instancia->NumNos  ) ;
                                 return NULL ;
                              }
                           } else {
                              printf( "Formato de especificacao de coordenadas de nos invalida [leitura do %do. no].\n", No  ) ;
                              return NULL ;
                           }
                        }

                        // Verifica se todos os ns foram lidos.
                        if ( No < Instancia->NumNos ) {
                           printf( "Foram lidos menos nos do que o especificado [Especificado: %d, Lidos: %d.\n", Instancia->NumNos, No ) ;
                           return NULL ;
                        }

                        break ;
                     /////////////////////////////////////////////////////////////
                     case SECAO_DEPOT:
                        Secoes = Secoes | SECAO_DEPOT_FLAG ; 
                        NosDepot = 0;
                        do {
                           CamposLidos = fscanf( Arquivo, "%d", &Inteiro ) ;
                           if ( CamposLidos > 0 ) {
                              // Ignorada ;
                              NosDepot++;
                           } else {
                              printf( "Secao de nos 'depot' nao e terminada por um -1.\n" ) ;
                              return NULL ;
                           }
                        } while ( Inteiro != -1 ) ;
                        break ;
                     /////////////////////////////////////////////////////////////
                     case SECAO_DEMANDA:
                        Secoes = Secoes | SECAO_DEMANDA_FLAG ; 
                        i = Instancia->NumNos + NosDepot ;
                        while ( i-- > 0 ) {
                           CamposLidos = fscanf( Arquivo, "%d %d", &Inteiro ) ;
                           if ( CamposLidos > 0 ) {
                              // Ignorada ;
                           } else {
                              printf( "Foram especificados menos nos na secao de demanda do que o exigido [nos=%d; depot=%d].\n", Instancia->NumNos, NosDepot ) ;
                              return NULL ;
                           }
                        } 
                        break ;
                     /////////////////////////////////////////////////////////////
                     case SECAO_ARESTAS:
                        Secoes = Secoes | SECAO_ARESTAS_FLAG ;

                        switch ( Espec[ FORMATO_ARESTA ] ) {

                        case LISTA_ADJACENCIAS:
                           do {
                              CamposLidos = fscanf( Arquivo, "%d", &i ) ;
                              if ( CamposLidos > 0 ) {
                                 if ( ( i > 0 ) && ( i <= Instancia->NumNos ) ) {
                                 
                                    do {
                                       CamposLidos = fscanf( Arquivo, "%d", &j ) ;
                                       if ( CamposLidos > 0 ) {
                                          if ( ( j > 0 ) && ( j <= Instancia->NumNos ) ) {

                                             if ( i < j ) Instancia->Grafo[ j ][ i ] = 1 ;
                                             else         Instancia->Grafo[ i ][ j ] = 1 ;

                                          } else {
                                             if ( j == -1 ) break ;
                                             else {
                                                printf( "Secao de arestas nao e terminada por um -1. [%d encontrado]\n", j ) ;
                                                return NULL ;
                                             }
                                          }
                                       } else {
                                          printf( "Secao de arestas nao e terminada por um -1.NULL" ) ;
                                          return NULL ;
                                       }
                                    } while ( 1 ) ;

                                 } else {
                                    if ( i == -1 ) break ;
                                    else {
                                       printf( "Secao de arestas nao e terminada por um -1. [%d encontrado]\n", i ) ;
                                       return NULL ;
                                    }
                                 }
                              } else {
                                 printf( "Secao de arestas nao e terminada por um -1.\n" ) ;
                                 return NULL ;
                              }
                           } while ( 1 ) ;
                           break ;

                        case LISTA_ARESTAS:
                           do {
                              CamposLidos = fscanf( Arquivo, "%d", &i ) ;
                              if ( CamposLidos > 0 ) {
                                 if ( ( i > 0 ) && ( i <= Instancia->NumNos ) ) {

                                    CamposLidos = fscanf( Arquivo, "%d", &j ) ;
                                    if ( CamposLidos > 0 ) {
                                       if ( ( j > 0 ) && ( j <= Instancia->NumNos ) ) {
                                          if ( i < j ) Instancia->Grafo[ j-1 ][ i-1 ] = 1 ;
                                          else         Instancia->Grafo[ i-1 ][ j-1 ] = 1 ;
                                       } else {
                                          printf( "Secao de arestas terminada antes do esperado. [%d encontrado]\n", j ) ;
                                          return NULL ;
                                       }
                                    } else {
                                       printf( "Secao de arestas nao e terminada por um -1.\n" ) ;
                                       return NULL ;
                                    }

                                 } else {
                                    if ( i == -1 ) break ;
                                    else {
                                       printf( "Secao de arestas nao e terminada por um -1. [%d encontrado]\n", i ) ;
                                       return NULL ;
                                    }
                                 }
                              } else {
                                 printf( "Secao de arestas nao e terminada por um -1.\n" ) ;
                                 return NULL ;
                              }
                           } while ( 1 ) ;
                           break ;

                        default:
                           printf( "Secao de arestas nao foi especificada, mas a secao foi encontrada.\n" ) ;
                           return NULL ;
                           break ;

                        }
                        break ;
                     /////////////////////////////////////////////////////////////
                     case SECAO_ARESTAS_FIXAS:
                        Secoes = Secoes | SECAO_ARESTAS_FIXAS_FLAG ;

                        do {
                           CamposLidos = fscanf( Arquivo, "%d", &i ) ;
                           if ( CamposLidos > 0 ) {
                              if ( ( i > 0 ) && ( i <= Instancia->NumNos ) ) {

                                 ///////////////
                                 // Ignorada. //
                                 ///////////////

                              } else {
                                 if ( i == -1 ) break ;
                                 else {
                                    printf( "Secao de arestas fixas nao e terminada por um -1. [%d encontrado]\n", i ) ;
                                    return NULL ;
                                 }
                              }
                           } else {
                              printf( "Secao de arestas fixas nao e terminada por um -1.\n" ) ;
                              return NULL ;
                           }
                        } while ( 1 ) ;
                        break ;
                     /////////////////////////////////////////////////////////////
                     case SECAO_EXIBICAO:
                        Secoes = Secoes | SECAO_EXIBICAO_FLAG ;

                        i = Instancia->NumNos ;
                        while ( i-- > 0 ) {
                           CamposLidos = fscanf( Arquivo, "%d %lf %lf", &Inteiro, &Real, &Real ) ;
                           if ( CamposLidos > 0 ) {
                              // Ignorada ;
                           } else {
                              printf( "Foram especificados menos nos na secao de exibicao do que o exigido [nos=%d; depot=%d].\n", Instancia->NumNos, NosDepot ) ;
                              return NULL ;
                           }
                        } 
                        break ;
                     /////////////////////////////////////////////////////////////
                     case SECAO_ROTA:
                        Secoes = Secoes | SECAO_ROTA_FLAG ;

                        Instancia->NumRot = 0 ;
                        Instancia->Rotas = NULL ;
                        do {
                           CamposLidos = fscanf( Arquivo, "%d", &i ) ;
                           if ( CamposLidos > 0 ) {
                              if ( ( i > 0 ) && ( i <= Instancia->NumNos ) ) {

                                 Instancia->NumRot++ ;
                                 Instancia->Rotas = ( tpNo** )realloc( Instancia->Rotas, sizeof( tpNo* ) * Instancia->NumRot ) ;
                                 Instancia->Rotas[ Instancia->NumRot - 1 ] = ( tpNo* )malloc( sizeof( tpNo ) * Instancia->NumNos ) ;
                                 No = i ;
                                 do {
                                    CamposLidos = fscanf( Arquivo, "%d", &j ) ;
                                    if ( CamposLidos > 0 ) {
                                       if ( ( j > 0 ) && ( j <= Instancia->NumNos ) ) {

                                          Instancia->Rotas[ Instancia->NumRot - 1 ][ i-1 ] = j-1 ;
                                          i = j ;

                                       } else {
                                          if ( j == -1 ) {
                                             Instancia->Rotas[ Instancia->NumRot - 1 ][ i-1 ] = No-1 ;
                                             break ;
                                          }
                                          else {
                                             printf( "Secao de rotas nao e terminada por um -1. [%d encontrado]\n", j ) ;
                                             return NULL ;
                                          }
                                       }
                                    } else {
                                       printf( "Secao de rotas nao e terminada por um -1.\n" ) ;
                                       return NULL ;
                                    }
                                 } while ( 1 ) ;

                              } else {
                                 if ( i == -1 ) break ;
                                 else {
                                    printf( "Secao de rotas nao e terminada por um -1. [%d encontrado]\n", i ) ;
                                    return NULL ;
                                 }
                              }
                           } else {
                              printf( "Secao de rotas nao e terminada por um -1.\n" ) ;
                              return NULL ;
                           }
                        } while ( 1 ) ;
                        break ;
                     /////////////////////////////////////////////////////////////
                     case SECAO_CUSTOS:
                        Secoes = Secoes | SECAO_CUSTOS_FLAG ;

                        switch( Espec[ FORMATO_CUSTO ] ) {
                        case FUNCAO_CUSTO:
                           printf( "A instancia especifica uma funcao para o calculo das distancias, mas foi encontrada uma secao de custos.\n" ) ;
                           return NULL ;
                           break ;
                        case MATRIZ_COMPLETA:
                           for( i = 0 ; i < Instancia->NumNos ; i++ ) {
                              for( j = 0 ; j < Instancia->NumNos ; j++ ) {
                                 CamposLidos = fscanf( Arquivo, "%d", &Custo ) ;
                                 if ( CamposLidos > 0 ) {
                                    if ( i >= j ) Instancia->Grafo[ i ][ j ] = Custo ;
                                 }
                              }
                           }
                           break ;
                        case MATRIZ_INF_COL:
                        case MATRIZ_SUP_LIN:
                           for( i = 0 ; i < Instancia->NumNos ; i++ ) {
                              Instancia->Grafo[ i ][ i ] = 0 ;
                              for( j = i + 1 ; j < Instancia->NumNos ; j++ ) {
                                 CamposLidos = fscanf( Arquivo, "%d", &Custo ) ;
                                 if ( CamposLidos > 0 ) {
                                    Instancia->Grafo[ j ][ i ] = Custo ;
                                 }
                              }
                           }
                           break ;
                        case MATRIZ_SUP_COL:
                        case MATRIZ_INF_LIN:
                           for( i = 0 ; i < Instancia->NumNos ; i++ ) {
                              Instancia->Grafo[ i ][ i ] = 0 ;
                              for( j = 0 ; j < i ; j++ ) {
                                 CamposLidos = fscanf( Arquivo, "%d", &Custo ) ;
                                 if ( CamposLidos > 0 ) {
                                    Instancia->Grafo[ i ][ j ] = Custo ;
                                 }
                              }
                           }
                           break ;
                        case MATRIZ_INF_DIAG_COL:
                        case MATRIZ_SUP_DIAG_LIN:
                           for( i = 0 ; i < Instancia->NumNos ; i++ ) {
                              for( j = i ; j < Instancia->NumNos ; j++ ) {
                                 CamposLidos = fscanf( Arquivo, "%d", &Custo ) ;
                                 if ( CamposLidos > 0 ) {
                                    Instancia->Grafo[ j ][ i ] = Custo ;
                                 }
                              }
                           }
                           break ;
                        case MATRIZ_SUP_DIAG_COL:
                        case MATRIZ_INF_DIAG_LIN:
                           for( i = 0 ; i < Instancia->NumNos ; i++ ) {
                              Instancia->Grafo[ i ][ i ] = 0 ;
                              for( j = 0 ; j <= i ; j++ ) {
                                 CamposLidos = fscanf( Arquivo, "%d", &Custo ) ;
                                 if ( CamposLidos > 0 ) {
                                    Instancia->Grafo[ i ][ j ] = Custo ;
                                 }
                              }
                           }
                        }
                        break ;
                     /////////////////////////////////////////////////////////////
                     case FIM:
                        Secoes = Secoes | FIM_FLAG ;
                        CamposLidos = 0 ;
                        break ;
                     /////////////////////////////////////////////////////////////
                     default:
                        printf( "Matriz de KeyWords esta incorreta. A KeyWord '%s' foi encontrada na tabela da posicao %d, mas nenhum tratamento adequado foi encontado.\n", Texto, KeyWord ) ;
                        return NULL ;
                     }
                  
                  }
                  break ;
               }
            }

            if ( KeyWord == NUM_KW ) {
               printf( "O campo '%s' nao corrresponde a nenhuma KeyWord do modelo de instancias da TSP-Lib.\n", Texto ) ;
               break ;
            }

         }

      } while ( CamposLidos > 0 ) ;

      if ( Espec[ TIPO_CUSTO ] == EXCPLICITO ) {
         if ( !( Secoes & SECAO_CUSTOS_FLAG ) || ( Secoes & SECAO_ARESTAS_FLAG ) ) {
            printf( "A instancia foi especificada com custos explicitos, porem nao foi encotrada a secao de custos ou foi encontrada um secao de arestas.\n" ) ;
            return NULL ;
         }
      } else {
         CalcularCustos( Instancia, Coords, Espec[ TIPO_CUSTO ] ) ;
      }

   } else {

      printf( "Nao foi possivel ler o arquivo %s\n", Caminho ) ;

   }

   free( Valores[ TIPO ] ) ;
   free( Valores[ TIPO_CUSTO ] ) ;
   free( Valores[ FORMATO_CUSTO ] ) ;
   free( Valores[ FORMATO_ARESTA ] ) ;
   free( Valores[ COORDENADAS ] ) ;
   free( Valores[ TIPO_EXIBICAO ] ) ;

   // Desaloca espao para armazenar as coordenadas.
   if ( ( Secoes & SECAO_COORD_NOS_FLAG ) && ( NumCoord > 0 ) ) {
      for ( i = 0 ; i < NumCoord ; i++ ) {
         free( Coords[ i ] ) ;
      }
      free( Coords ) ;
   }

   fclose( Arquivo ) ;
/*
   KeyWords[ NOME           ] = "NAME : %s\n"               ;
   KeyWords[ TIPO           ] = "TYPE : %d\n"               ;
   KeyWords[ COMENTARIO     ] = "COMMENT : %s\n"            ;
   KeyWords[ DIMENSAO       ] = "DIMENSION : %d\n"          ;
   KeyWords[ CAPACIDADE     ] = "CAPACITY : %d\n"           ;
   KeyWords[ TIPO_CUSTO     ] = "EDGE_WEIGHT_TYPE : %d\n"   ;
   KeyWords[ FORMATO_CUSTO  ] = "EDGE_WEIGHT_FORMAT : %d\n" ;
   KeyWords[ FORMATO_ARESTA ] = "EDGE_DATA_FORMAT : %d\n"   ;
   KeyWords[ COORDENADAS    ] = "NODE_COORD_TYPE : %d\n"    ;
   KeyWords[ TIPO_EXIBICAO  ] = "DISPLAY_DATA_TYPE : %d\n"  ;

   printf( KeyWords[ NOME ], Instancia->Nome ) ;
   printf( KeyWords[ COMENTARIO ], Instancia->Comentario ) ;
   printf( KeyWords[ DIMENSAO ], Instancia->NumNos ) ;
   printf( KeyWords[ CAPACIDADE ], -1 ) ;
   for ( KeyWord = 0 ; KeyWord < NUM_KW_COM_VAL ; KeyWord++ ) {
      printf( KeyWords[ KeyWord ], Espec[ KeyWord ] ) ;
   }
//*/
   return Instancia ;
}

tpCusto VerificarRotaCanonica( tpGrafo Grafo, tpNo NumNos ) {

   int     i         ;
   tpCusto Custo = 0 ;

   for( i = 1 ; i < NumNos ; i++ ) {
      Custo += Grafo[ i ][ i-1 ] ;
   }

   return Custo + Grafo[ NumNos - 1 ][ 0 ] ;

}

tpCusto VerificarRota( tpGrafo Grafo, tpNo* Rota ) {

   int     i         ;
   tpCusto Custo = 0 ;

   i = Rota[ 0 ] ;
   do {
      Custo += PESO( Grafo, i, Rota[ i ] ) ;
      i = Rota[ i ] ;
   } while ( i != Rota[ 0 ] ) ;

   return Custo ;

}